import pytest

from ansible.errors import AnsibleError
from conftest import ALL_HOSTS
from conftest import NEGATIVE_HOST_PATTERNS
from conftest import NEGATIVE_HOST_SLICES
from conftest import POSITIVE_HOST_PATTERNS
from conftest import POSITIVE_HOST_SLICES


pytestmark = [
    pytest.mark.unit,
]


def test_len(hosts):
    assert len(hosts) == len(ALL_HOSTS)


def test_keys(hosts):
    sorted_keys = hosts.keys()
    sorted_keys.sort()
    assert sorted_keys == ALL_HOSTS


@pytest.mark.parametrize(("host_pattern", "num_hosts"), POSITIVE_HOST_PATTERNS)
def test_contains(host_pattern, num_hosts, hosts):
    assert host_pattern in hosts, f"{host_pattern} not in hosts"


@pytest.mark.parametrize(("host_pattern", "num_hosts"), NEGATIVE_HOST_PATTERNS)
def test_not_contains(host_pattern, num_hosts, hosts):
    assert host_pattern not in hosts


@pytest.mark.parametrize(("host_pattern", "num_hosts"), POSITIVE_HOST_PATTERNS)
def test_getitem(host_pattern, num_hosts, hosts):
    assert hosts[host_pattern]


@pytest.mark.parametrize(("host_pattern", "num_hosts"), NEGATIVE_HOST_PATTERNS)
def test_not_getitem(host_pattern, num_hosts, hosts):
    with pytest.raises(KeyError):
        assert hosts[host_pattern]


@pytest.mark.parametrize(("host_pattern", "num_hosts"), POSITIVE_HOST_PATTERNS)
def test_getattr(host_pattern, num_hosts, hosts):
    assert hasattr(hosts, host_pattern)


@pytest.mark.parametrize(("host_slice", "num_hosts"), POSITIVE_HOST_SLICES)
def test_slice(host_slice, num_hosts, hosts):
    assert (
        len(hosts[host_slice]) == num_hosts
    ), f"{len(hosts[host_slice])} != {num_hosts} for {host_slice}"


# pylint: disable=pointless-statement
@pytest.mark.parametrize("host_slice", NEGATIVE_HOST_SLICES)
def test_not_slice(host_slice, hosts):
    with pytest.raises(KeyError):
        hosts[host_slice]


@pytest.mark.parametrize(("host_pattern", "num_hosts"), NEGATIVE_HOST_PATTERNS)
def test_not_getattr(host_pattern, num_hosts, hosts):
    assert not hasattr(hosts, host_pattern)
    with pytest.raises(AttributeError):
        getattr(hosts, host_pattern)


# This should probably be made more generic for all options (and moved elsewhere)
@pytest.mark.ansible_v1_xfail(raises=AnsibleError)
def test_defaults(request):
    from ansible.constants import DEFAULT_TRANSPORT

    plugin = request.config.pluginmanager.getplugin("ansible")
    hosts = plugin.initialize(config=request.config, request=request)

    # from pytest_ansible.host_manager import get_host_manager
    # hosts = get_host_manager(inventory='unknown.example.com,')
    assert "connection" in hosts.options
    assert hosts.options["connection"] == DEFAULT_TRANSPORT



"""PyTest Ansible Plugin."""

import ansible
import ansible.constants
import ansible.errors
import ansible.utils
import pytest

from ansible.plugins.loader import become_loader

from pytest_ansible.fixtures import ansible_adhoc
from pytest_ansible.fixtures import ansible_facts
from pytest_ansible.fixtures import ansible_module
from pytest_ansible.fixtures import localhost
from pytest_ansible.host_manager import get_host_manager


# Silence linters for imported fixtures
# pylint: disable=pointless-statement
(ansible_adhoc, ansible_module, ansible_facts, localhost)


def become_methods():
    """Return string list of become methods available to ansible."""
    if become_loader:
        return [method.name for method in become_loader.all()]
    return ansible.constants.BECOME_METHODS


def pytest_addoption(parser):
    """Add options to control ansible."""

    group = parser.getgroup("pytest-ansible")
    group.addoption(
        "--inventory",
        "--ansible-inventory",
        action="store",
        dest="ansible_inventory",
        default=ansible.constants.DEFAULT_HOST_LIST,
        metavar="ANSIBLE_INVENTORY",
        help="ansible inventory file URI (default: %(default)s)",
    )
    group.addoption(
        "--extra-inventory",
        "--ansible-extra-inventory",
        action="store",
        dest="ansible_extra_inventory",
        default=None,
        metavar="ANSIBLE_EXTRA_INVENTORY",
        help="ansible extra inventory file URI (default: %(default)s)",
    )
    group.addoption(
        "--host-pattern",
        "--ansible-host-pattern",
        action="store",
        dest="ansible_host_pattern",
        default=None,
        metavar="ANSIBLE_HOST_PATTERN",
        help="ansible host pattern (default: %(default)s)",
    )
    group.addoption(
        "--limit",
        "--ansible-limit",
        action="store",
        dest="ansible_subset",
        default=ansible.constants.DEFAULT_SUBSET,
        metavar="ANSIBLE_SUBSET",
        help="further limit selected hosts to an additional pattern",
    )
    group.addoption(
        "--connection",
        "--ansible-connection",
        action="store",
        dest="ansible_connection",
        default=ansible.constants.DEFAULT_TRANSPORT,
        help="connection type to use (default: %(default)s)",
    )
    group.addoption(
        "--user",
        "--ansible-user",
        action="store",
        dest="ansible_user",
        default=ansible.constants.DEFAULT_REMOTE_USER,
        help="connect as this user (default: %(default)s)",
    )
    group.addoption(
        "--check",
        "--ansible-check",
        action="store_true",
        dest="ansible_check",
        default=False,
        help="don't make any changes; instead, try to predict some of the changes that may occur",
    )
    group.addoption(
        "--module-path",
        "--ansible-module-path",
        action="store",
        dest="ansible_module_path",
        default=ansible.constants.DEFAULT_MODULE_PATH,
        help="specify path(s) to module library (default: %(default)s)",
    )

    # become privilege escalation
    group.addoption(
        "--become",
        "--ansible-become",
        action="store_true",
        dest="ansible_become",
        default=ansible.constants.DEFAULT_BECOME,
        help="run operations with become, nopasswd implied (default: %(default)s)",
    )
    group.addoption(
        "--become-method",
        "--ansible-become-method",
        action="store",
        dest="ansible_become_method",
        default=ansible.constants.DEFAULT_BECOME_METHOD,
        help=f"privilege escalation method to use (default: %(default)s), valid choices: [ {' | '.join(become_methods())} ]",
    )

    group.addoption(
        "--become-user",
        "--ansible-become-user",
        action="store",
        dest="ansible_become_user",
        default=ansible.constants.DEFAULT_BECOME_USER,
        help="run operations as this user (default: %(default)s)",
    )
    group.addoption(
        "--ask-become-pass",
        "--ansible-ask-become-pass",
        action="store",
        dest="ansible_ask_become_pass",
        default=ansible.constants.DEFAULT_BECOME_ASK_PASS,
        help="ask for privilege escalation password (default: %(default)s)",
    )

    # Add github marker to --help
    parser.addini("ansible", "Ansible integration", "args")


def pytest_configure(config):
    """Validate --ansible-* parameters."""

    config.addinivalue_line("markers", "ansible(**kwargs): Ansible integration")

    # Enable connection debugging
    if config.option.verbose > 0:
        if hasattr(ansible.utils, "VERBOSITY"):
            ansible.utils.VERBOSITY = int(config.option.verbose)
        else:
            from ansible.utils.display import Display

            display = Display()
            display.verbosity = int(config.option.verbose)

    assert config.pluginmanager.register(PyTestAnsiblePlugin(config), "ansible")


def pytest_generate_tests(metafunc):
    """Generate tests when specific `ansible_*` fixtures are used by tests."""

    if "ansible_host" in metafunc.fixturenames:
        # assert required --ansible-* parameters were used
        PyTestAnsiblePlugin.assert_required_ansible_parameters(metafunc.config)
        try:
            plugin = metafunc.config.pluginmanager.getplugin("ansible")
            hosts = plugin.initialize(
                config=plugin.config,
                pattern=metafunc.config.getoption("ansible_host_pattern"),
            )
        except ansible.errors.AnsibleError as e:
            raise pytest.UsageError(e)
        # Return the host name as a string
        # metafunc.parametrize("ansible_host", hosts.keys())
        # Return a HostManager instance where pattern=host (e.g. ansible_host.all.shell('date'))
        # metafunc.parametrize("ansible_host", iter(plugin.initialize(config=plugin.config, pattern=h) for h in
        #                                           hosts.keys()))
        # Return a ModuleDispatcher instance representing `host` (e.g. ansible_host.shell('date'))
        metafunc.parametrize("ansible_host", iter(hosts[h] for h in hosts.keys()))

    if "ansible_group" in metafunc.fixturenames:
        # assert required --ansible-* parameters were used
        PyTestAnsiblePlugin.assert_required_ansible_parameters(metafunc.config)
        try:
            plugin = metafunc.config.pluginmanager.getplugin("ansible")
            hosts = plugin.initialize(
                config=plugin.config,
                pattern=metafunc.config.getoption("ansible_host_pattern"),
            )
        except ansible.errors.AnsibleError as e:
            raise pytest.UsageError(e)
        # FIXME: Eeew, this shouldn't be interfacing with `hosts.options`
        groups = hosts.options["inventory_manager"].list_groups()
        extra_groups = hosts.get_extra_inventory_groups()
        # Return the group name as a string
        # metafunc.parametrize("ansible_group", groups)
        # Return a ModuleDispatcher instance representing the group (e.g. ansible_group.shell('date'))
        metafunc.parametrize("ansible_group", iter(hosts[g] for g in groups))
        metafunc.parametrize("ansible_group", iter(hosts[g] for g in extra_groups))


class PyTestAnsiblePlugin:

    """Ansible PyTest Plugin Class."""

    def __init__(self, config):
        """Initialize plugin."""
        self.config = config

    def pytest_report_header(self, config, startdir):
        """Return the version of ansible."""
        return f"ansible: {ansible.__version__}"

    def pytest_collection_modifyitems(self, session, config, items):
        """Validate --ansible-* parameters."""

        uses_ansible_fixtures = False
        for item in items:
            if not hasattr(item, "fixturenames"):
                continue
            if any(fixture.startswith("ansible_") for fixture in item.fixturenames):
                # TODO - ignore if they are using a marker
                # marker = item.get_marker('ansible')
                # if marker and 'inventory' in marker.kwargs:
                uses_ansible_fixtures = True
                break

        if uses_ansible_fixtures:
            # assert required --ansible-* parameters were used
            self.assert_required_ansible_parameters(config)

    def _load_ansible_config(self, config):
        """Load ansible configuration from command-line."""
        option_names = [
            "ansible_inventory",
            "ansible_extra_inventory",
            "ansible_host_pattern",
            "ansible_connection",
            "ansible_user",
            "ansible_module_path",
            "ansible_become",
            "ansible_become_method",
            "ansible_become_user",
            "ansible_ask_become_pass",
            "ansible_subset",
        ]

        kwargs = {}

        # Load command-line supplied values
        for key in option_names:
            short_key = key[8:]
            kwargs[short_key] = config.getoption(key)

        # normalize ansible.ansible_become options
        kwargs["become"] = kwargs.get("become") or ansible.constants.DEFAULT_BECOME
        kwargs["become_user"] = (
            kwargs.get("become_user") or ansible.constants.DEFAULT_BECOME_USER
        )
        kwargs["ask_become_pass"] = (
            kwargs.get("ask_become_pass") or ansible.constants.DEFAULT_BECOME_ASK_PASS
        )

        return kwargs

    def _load_request_config(self, request):
        """Load ansible configuration from decorator kwargs."""
        kwargs = {}

        # Override options from @pytest.mark.ansible
        marker = request.node.get_closest_marker("ansible")
        if marker:
            kwargs = marker.kwargs

        return kwargs

    def initialize(self, config=None, request=None, **kwargs):
        """Return an initialized Ansible Host Manager instance."""
        ansible_cfg = {}
        # merge command-line configuration options
        if config is not None:
            ansible_cfg.update(self._load_ansible_config(config))
        # merge pytest request configuration options
        if request is not None:
            ansible_cfg.update(self._load_request_config(request))
        # merge in provided kwargs
        ansible_cfg.update(kwargs)
        return get_host_manager(**ansible_cfg)

    @staticmethod
    def assert_required_ansible_parameters(config):
        """Assert whether the required --ansible-* parameters were provided."""
        errors = []

        # Verify --ansible-host-pattern was provided
        ansible_hostname = config.getoption("ansible_host_pattern")
        if ansible_hostname is None or ansible_hostname == "":
            errors.append(
                "Missing required parameter --ansible-host-pattern/--host-pattern"
            )

        # NOTE: I don't think this will ever catch issues since ansible_inventory
        # defaults to '/etc/ansible/hosts'
        # Verify --ansible-inventory was provided
        ansible_inventory = config.getoption("ansible_inventory")
        if ansible_inventory is None or ansible_inventory == "":
            errors.append(
                "Unable to find an inventory file, specify one with the --ansible-inventory/--inventory "
                "parameter."
            )

        if errors:
            raise pytest.UsageError(*errors)



from ansible.inventory.manager import InventoryManager
from ansible.parsing.dataloader import DataLoader
from ansible.vars.manager import VariableManager

from pytest_ansible.host_manager import BaseHostManager
from pytest_ansible.module_dispatcher.v29 import ModuleDispatcherV29


class HostManagerV29(BaseHostManager):
    """Fixme."""

    def __init__(self, *args, **kwargs):
        """Fixme."""
        super().__init__(*args, **kwargs)
        self._dispatcher = ModuleDispatcherV29

    def initialize_inventory(self):
        self.options["loader"] = DataLoader()
        self.options["inventory_manager"] = InventoryManager(
            loader=self.options["loader"], sources=self.options["inventory"]
        )
        self.options["variable_manager"] = VariableManager(
            loader=self.options["loader"], inventory=self.options["inventory_manager"]
        )
        if "extra_inventory" in self.options:
            self.options["extra_loader"] = DataLoader()
            self.options["extra_inventory_manager"] = InventoryManager(
                loader=self.options["extra_loader"],
                sources=self.options["extra_inventory"],
            )
            self.options["extra_variable_manager"] = VariableManager(
                loader=self.options["extra_loader"],
                inventory=self.options["extra_inventory_manager"],
            )



import pytest

from pytest_ansible.has_version import has_ansible_v1
from pytest_ansible.has_version import has_ansible_v24


try:
    from ansible.utils import context_objects as co
except ImportError:
    # if it does not exist because of old version of ansible, we don't need it
    co = None


pytest_plugins = ("pytester",)


ALL_HOSTS = ["another_host", "localhost", "yet_another_host"]

POSITIVE_HOST_PATTERNS = [
    ("all", 3),
    ("*", 3),
    ("localhost", 1),
    ("local*", 1),
    ("local*,&*host", 1),
    ("!localhost", 2),
    ("all[0]", 1),
    ("all[-1]", 1),
    pytest.param("*[0-1]", 1, marks=pytest.mark.requires_ansible_v1()),
    pytest.param("*[0-1]", 2, marks=pytest.mark.requires_ansible_v2()),
    pytest.param(
        "*[0:1]", 2, marks=pytest.mark.requires_ansible_v2()
    ),  # this is confusing, but how host slicing works on v2
    pytest.param("*[0:]", 3, marks=pytest.mark.requires_ansible_v2()),
]

NEGATIVE_HOST_PATTERNS = [
    ("none", 0),
    ("all[8:]", 0),
]

POSITIVE_HOST_SLICES = [
    (slice(0, 0), 1),
    pytest.param(slice(0, 1), 1, marks=pytest.mark.requires_ansible_v1()),
    pytest.param(slice(0, 1), 2, marks=pytest.mark.requires_ansible_v2()),
    pytest.param(slice(0, 2), 2, marks=pytest.mark.requires_ansible_v1()),
    pytest.param(slice(0, 2), 3, marks=pytest.mark.requires_ansible_v2()),
    (slice(0), 1),
    (slice(1), 1),
    (slice(2), 1),
    (slice(3), 1),
]

NEGATIVE_HOST_SLICES = [
    slice(None),
    slice(-1),
]


def pytest_runtest_setup(item):
    # Conditionally skip tests that are pinned to a specific ansible version
    if isinstance(item, pytest.Function):
        # conditionally skip
        if item.get_closest_marker("requires_ansible_v1") and not has_ansible_v1:
            pytest.skip("requires < ansible-2.*")
        if item.get_closest_marker("requires_ansible_v2") and has_ansible_v1:
            pytest.skip("requires >= ansible-2.*")
        if item.get_closest_marker("requires_ansible_v24") and not has_ansible_v24:
            pytest.skip("requires >= ansible-2.4.*")
        if item.get_closest_marker("requires_ansible_v28") and not has_ansible_v24:
            pytest.skip("requires >= ansible-2.8.*")

        # conditionally xfail
        mark = item.get_closest_marker("ansible_v1_xfail")
        if mark and has_ansible_v1:
            item.add_marker(
                pytest.mark.xfail(
                    reason="expected failure on < ansible-2.*",
                    raises=mark.kwargs.get("raises"),
                )
            )

        mark = item.get_closest_marker("ansible_v2_xfail")
        if mark and not has_ansible_v1:
            item.add_marker(
                pytest.xfail(
                    reason="expected failure on >= ansible-2.*",
                    raises=mark.kwargs.get("raises"),
                )
            )


# pylint: disable=too-few-public-methods
class PyTestOption:
    """Helper class that provides methods for creating and managing an inventory file."""

    def __init__(self, config, testdir):
        self.config = config

        # Create inventory file
        self.inventory = testdir.makefile(
            ".ini",
            inventory="""
            [local]

            [local:children]
            reachable

            [reachable]
            localhost ansible_connection=local ansible_python_interpreter='/usr/bin/env python'
            127.0.0.2 ansible_connection=local ansible_python_interpreter='/usr/bin/env python'
            127.0.0.3 ansible_connection=local ansible_python_interpreter='/usr/bin/env python'
            127.0.0.4 ansible_connection=local ansible_python_interpreter='/usr/bin/env python'
            127.0.0.5 ansible_connection=local ansible_python_interpreter='/usr/bin/env python'

            [unreachable]
            unreachable-host-1.example.com
            unreachable-host-2.example.com
            unreachable-host-3.example.com
        """,
        )

        # Create ansible.cfg file
        # self.ansible_cfg = testdir.makefile('.cfg', ansible='''[ssh_connection]\ncontrol_path = %(directory)s/%%h-%%r''')

    @property
    def args(self):
        args = []
        args.append("--tb")
        args.append("native")
        return args


@pytest.fixture(autouse=True)
def _clear_global_context():
    # Reset the stored command line args
    # if context object does not exist because of old version of ansible, we don't need it
    if co is not None:
        co.GlobalCLIArgs._Singleton__instance = None


@pytest.fixture()
def option(request, testdir):
    """Returns an instance of PyTestOption to help tests pass parameters and
    use a common inventory file.
    """
    return PyTestOption(request.config, testdir)


@pytest.fixture()
def hosts():
    from pytest_ansible.host_manager import get_host_manager

    return get_host_manager(inventory=",".join(ALL_HOSTS), connection="local")



from ansible.inventory import Inventory
from ansible.parsing.dataloader import DataLoader
from ansible.vars import VariableManager

from pytest_ansible.host_manager import BaseHostManager
from pytest_ansible.module_dispatcher.v2 import ModuleDispatcherV2


class HostManagerV2(BaseHostManager):

    """Fixme."""

    def __init__(self, *args, **kwargs):
        """Fixme."""
        super().__init__(*args, **kwargs)
        self._dispatcher = ModuleDispatcherV2

    def initialize_inventory(self):
        self.options["loader"] = DataLoader()
        self.options["variable_manager"] = VariableManager()
        self.options["inventory_manager"] = Inventory(
            loader=self.options["loader"],
            variable_manager=self.options["variable_manager"],
            host_list=self.options["inventory"],
        )
        self.options["variable_manager"].set_inventory(
            self.options["inventory_manager"]
        )



import warnings

import ansible
import ansible.constants
import ansible.errors
import ansible.utils

from ansible.runner import Runner

from pytest_ansible.errors import AnsibleConnectionFailure
from pytest_ansible.has_version import has_ansible_v1
from pytest_ansible.module_dispatcher import BaseModuleDispatcher
from pytest_ansible.results import AdHocResult


if not has_ansible_v1:
    raise ImportError("Only supported with ansible < 2.0")


class ModuleDispatcherV1(BaseModuleDispatcher):

    """Pass."""

    required_kwargs = ("inventory", "inventory_manager", "host_pattern")

    def has_module(self, name):
        # Make sure we parse module_path and pass it to the loader,
        # otherwise, only built-in modules will work.
        if "module_path" in self.options:
            paths = self.options["module_path"]
            if isinstance(paths, (list, tuple, set)):
                for path in paths:
                    ansible.utils.module_finder.add_directory(path)
            else:
                ansible.utils.module_finder.add_directory(paths)

        return ansible.utils.module_finder.has_plugin(name)

    def _run(self, *module_args, **complex_args):
        """Execute an ansible adhoc command returning the results in a AdHocResult object."""
        # Assemble module argument string
        if module_args:
            complex_args.update({"_raw_params": " ".join(module_args)})
        else:
            module_args = " ".join(module_args)

        # Assert hosts matching the provided pattern exist
        hosts = self.options["inventory_manager"].list_hosts()
        no_hosts = False
        if len(hosts) == 0:
            no_hosts = True
            warnings.warn("provided hosts list is empty, only localhost is available")

        self.options["inventory_manager"].subset(self.options.get("subset"))
        hosts = self.options["inventory_manager"].list_hosts(
            self.options["host_pattern"]
        )
        if len(hosts) == 0 and not no_hosts:
            raise ansible.errors.AnsibleError(
                "Specified hosts and/or --limit does not match any hosts"
            )

        # Build module runner object
        kwargs = {
            "inventory": self.options.get("inventory_manager"),
            "pattern": self.options.get("host_pattern"),
            "module_name": self.options.get("module_name"),
            "module_args": module_args,
            "complex_args": complex_args,
            "transport": self.options.get("connection"),
            "remote_user": self.options.get("user"),
            "module_path": self.options.get("module_path"),
            "become": self.options.get("become"),
            "become_method": self.options.get("become_method"),
            "become_user": self.options.get("become_user"),
        }

        # Run the module
        runner = Runner(**kwargs)
        results = runner.run()

        if "dark" in results and results["dark"]:
            raise AnsibleConnectionFailure(
                "Host unreachable", dark=results["dark"], contacted=results["contacted"]
            )

        # Success!
        return AdHocResult(contacted=results["contacted"])



from ansible.inventory.manager import InventoryManager
from ansible.parsing.dataloader import DataLoader
from ansible.vars.manager import VariableManager

from pytest_ansible.host_manager import BaseHostManager
from pytest_ansible.module_dispatcher.v28 import ModuleDispatcherV28


class HostManagerV28(BaseHostManager):

    """Fixme."""

    def __init__(self, *args, **kwargs):
        """Fixme."""
        super().__init__(*args, **kwargs)
        self._dispatcher = ModuleDispatcherV28

    def initialize_inventory(self):
        self.options["loader"] = DataLoader()
        self.options["inventory_manager"] = InventoryManager(
            loader=self.options["loader"], sources=self.options["inventory"]
        )
        self.options["variable_manager"] = VariableManager(
            loader=self.options["loader"], inventory=self.options["inventory_manager"]
        )
        # self.options['inventory_manager'].clear_caches()



from ansible.inventory import Inventory

from pytest_ansible.host_manager import BaseHostManager
from pytest_ansible.module_dispatcher.v1 import ModuleDispatcherV1


class HostManagerV1(BaseHostManager):

    """Fixme."""

    def __init__(self, *args, **kwargs):
        """Fixme."""
        super().__init__(*args, **kwargs)
        self._dispatcher = ModuleDispatcherV1

    def keys(self):
        return list(self.options["inventory_manager"].list_hosts())

    def initialize_inventory(self):
        self.options["inventory_manager"] = Inventory(self.options["inventory"])
        # self.options['inventory_manager'].subset(self.pattern)



import re
import sys

from unittest import mock

import ansible
import pytest

from pkg_resources import parse_version

from pytest_ansible.has_version import has_ansible_v28


# pylint: disable=unused-import
try:
    from _pytest.main import EXIT_INTERRUPTED  # type: ignore[attr-defined]
    from _pytest.main import EXIT_NOTESTSCOLLECTED  # type: ignore[attr-defined]
    from _pytest.main import EXIT_OK  # type: ignore[attr-defined]
    from _pytest.main import EXIT_TESTSFAILED  # type: ignore[attr-defined]
    from _pytest.main import EXIT_USAGEERROR  # type: ignore[attr-defined]
except ImportError:
    from _pytest.main import ExitCode

    EXIT_OK = ExitCode.OK
    EXIT_TESTSFAILED = ExitCode.TESTS_FAILED
    EXIT_USAGEERROR = ExitCode.USAGE_ERROR
    EXIT_INTERRUPTED = ExitCode.INTERRUPTED
    EXIT_NOTESTSCOLLECTED = ExitCode.NO_TESTS_COLLECTED

# pylint: disable=unused-import
if sys.version_info[0] == 2:
    import __builtin__ as builtins  # NOQA
else:
    import builtins  # NOQA


def test_plugin_help(testdir):
    """Verifies expected output from of py.test --help"""

    result = testdir.runpytest("--help")
    result.stdout.fnmatch_lines(
        [
            # Check for the github args section header
            "pytest-ansible:",
            # Check for the specific args
            "  --inventory=ANSIBLE_INVENTORY, --ansible-inventory=ANSIBLE_INVENTORY",
            "  --host-pattern=ANSIBLE_HOST_PATTERN, --ansible-host-pattern=ANSIBLE_HOST_PATTERN",
            "  --connection=ANSIBLE_CONNECTION, --ansible-connection=ANSIBLE_CONNECTION",
            "  --user=ANSIBLE_USER, --ansible-user=ANSIBLE_USER",
            "  --check, --ansible-check",
            "  --module-path=ANSIBLE_MODULE_PATH, --ansible-module-path=ANSIBLE_MODULE_PATH",
            "  --become, --ansible-become",
            "  --become-method=ANSIBLE_BECOME_METHOD, --ansible-become-method=ANSIBLE_BECOME_METHOD",
            "  --become-user=ANSIBLE_BECOME_USER, --ansible-become-user=ANSIBLE_BECOME_USER",
            "  --ask-become-pass=ANSIBLE_ASK_BECOME_PASS, --ansible-ask-become-pass=ANSIBLE_ASK_BECOME_PASS",
            # Check for the marker in --help
            "  ansible (args)*Ansible integration",
        ]
    )


def test_plugin_markers(testdir):
    """Verifies expected output from of py.test --markers"""

    result = testdir.runpytest("--markers")
    result.stdout.fnmatch_lines(
        [
            "@pytest.mark.ansible(*args): Ansible integration",
        ]
    )


def test_report_header(testdir, option):
    """Verify the expected ansible version in the pytest report header."""

    result = testdir.runpytest(*option.args)
    assert result.ret == EXIT_NOTESTSCOLLECTED
    result.stdout.fnmatch_lines([f"ansible: {ansible.__version__}"])


def test_params_not_required_when_not_using_fixture(testdir, option):
    """Verify the ansible parameters are not required if the fixture is not used."""

    src = """
        import pytest
        def test_func():
            assert True
    """
    testdir.makepyfile(src)
    result = testdir.runpytest(*option.args)
    assert result.ret == EXIT_OK


@pytest.mark.parametrize(
    "fixture_name",
    (
        "ansible_adhoc",
        "ansible_module",
        "ansible_facts",
    ),
)
def test_params_required_when_using_fixture(testdir, option, fixture_name):
    """Verify the ansible parameters are required if the fixture is used."""

    src = f"""
        import pytest
        def test_func({fixture_name}):
            {fixture_name}
        """

    testdir.makepyfile(src)
    result = testdir.runpytest(*option.args)
    assert result.ret == EXIT_USAGEERROR
    result.stderr.fnmatch_lines(
        [
            "ERROR: Missing required parameter --ansible-host-pattern/--host-pattern",
        ]
    )


@pytest.mark.parametrize(
    "required_value_parameter",
    (
        "--ansible-inventory",
        "--inventory",
        "--ansible-host-pattern",
        "--host-pattern",
        "--ansible-connection",
        "--connection",
        "--ansible-user",
        "--user",
        "--ansible-become-method",
        "--become-method",
        "--ansible-become-user",
        "--become-user",
        "--ansible-module-path",
        "--module-path",
    ),
)
def test_param_requires_value(testdir, required_value_parameter):
    """Verifies failure when not providing a value to a parameter that requires a value"""

    result = testdir.runpytest(*[required_value_parameter])
    assert result.ret == EXIT_USAGEERROR
    result.stderr.fnmatch_lines(
        [f"*: error: argument *{required_value_parameter}*: expected one argument"]
    )


def test_params_required_with_inventory_without_host_pattern(testdir, option):
    """Verify that a host pattern is required when an inventory is supplied."""
    src = """
        import pytest
        def test_func(ansible_module):
            assert True
    """
    testdir.makepyfile(src)
    result = testdir.runpytest(*option.args + ["--ansible-inventory", "local,"])
    assert result.ret == EXIT_USAGEERROR
    result.stderr.fnmatch_lines(
        [
            "ERROR: Missing required parameter --ansible-host-pattern/--host-pattern",
        ]
    )


@pytest.mark.requires_ansible_v1()
def test_params_required_with_bogus_inventory_v1(testdir, option):
    src = """
        import pytest
        def test_func(ansible_module):
            assert True
    """
    testdir.makepyfile(src)
    with mock.patch("os.path.exists", return_value=False) as mock_exists:
        result = testdir.runpytest(
            *["--ansible-inventory", "bogus", "--ansible-host-pattern", "all"]
        )

    # Assert py.test exit code
    assert result.ret == EXIT_TESTSFAILED

    # Assert expected error output
    result.stdout.fnmatch_lines(
        [
            "*Unable to find an inventory file, specify one with -i ?",
        ]
    )

    # Assert mock open called on provided file
    mock_exists.assert_any_call("bogus")


@pytest.mark.skipif(
    parse_version(ansible.__version__) < parse_version("2.0.0")
    or parse_version(ansible.__version__) >= parse_version("2.4.0"),
    reason="requires ansible >= 2.0 and < 2.4",
)
def test_params_required_with_bogus_inventory_v2(testdir, option, recwarn):
    src = """
        import pytest
        def test_func(ansible_module):
            with pytest.warns(UserWarning, match="provided hosts list is empty, only localhost is available"):
                ansible_module.ping()
    """
    testdir.makepyfile(src)

    with mock.patch(
        "ansible.parsing.dataloader.DataLoader.path_exists", return_value=False
    ) as mock_exists:
        result = testdir.runpytest(
            *[
                "-vvvvvs",
                "--ansible-inventory",
                "bogus",
                "--ansible-host-pattern",
                "all",
            ]
        )

    # Assert py.test exit code
    assert result.ret == EXIT_OK

    # Assert mock open called on provided file
    mock_exists.assert_any_call("bogus")


@pytest.mark.requires_ansible_v24()
@pytest.mark.skipif(has_ansible_v28, reason="requires ansible < 2.8")
def test_params_required_with_bogus_inventory_v24(testdir, option, recwarn):
    src = """
        import pytest
        def test_func(ansible_module):
            with pytest.warns(UserWarning) as record:
                ansible_module.ping()
            # Ensure at least one warning in the queue
            assert len(record) >= 1
            # Ensure the latest warning is the ansible localhost warning
            assert record[0].message.args[0] == "provided hosts list is empty, only localhost is available"

    """
    testdir.makepyfile(src)

    result = testdir.runpytest(
        *["-vvvvvs", "--ansible-inventory", "bogus", "--ansible-host-pattern", "all"]
    )

    # Assert py.test exit code
    assert result.ret == EXIT_OK

    # There appear to be '\n' newline characters within the output.  Using the join on errlines flattens the string for
    # easier comparison.
    assert re.search(
        r"Unable to parse .*/bogus as an inventory source", " ".join(result.errlines)
    )


@pytest.mark.requires_ansible_v1()
def test_params_required_without_inventory_with_host_pattern_v1(testdir, option):
    src = """
        import pytest
        def test_func(ansible_module):
            assert True
    """
    testdir.makepyfile(src)
    result = testdir.runpytest(*option.args + ["--ansible-host-pattern", "all"])
    assert result.ret == EXIT_TESTSFAILED
    result.stdout.fnmatch_lines(
        [
            "*Unable to find an inventory file, specify one with -i ?",
        ]
    )


@pytest.mark.requires_ansible_v2()
def test_params_required_without_inventory_with_host_pattern_v2(testdir, option):
    src = """
        import pytest
        def test_func(ansible_module):
            assert True
    """
    testdir.makepyfile(src)
    result = testdir.runpytest(*option.args + ["--ansible-host-pattern", "all"])
    assert result.ret == EXIT_OK

    # TODO - validate the following warning message
    # [WARNING]: provided hosts list is empty, only localhost is available
    # pylint: disable=using-constant-test
    if False:
        result.stderr.fnmatch_lines(
            [
                "*[WARNING]: Host file not found: /etc/ansible/hosts*",
                "*provided hosts list is empty, only localhost is available",
            ]
        )


def test_param_override_with_marker(testdir):
    src = """
        import pytest
        @pytest.mark.ansible(inventory='local,', connection='local', host_pattern='all')
        def test_func(ansible_module):
            ansible_module.ping()
    """
    testdir.makepyfile(src)
    result = testdir.runpytest(
        *[
            "-vvvvvs",
            "--tb",
            "native",
            "--ansible-inventory",
            "garbage,",
            "--ansible-host-pattern",
            "garbage",
            "--ansible-connection",
            "garbage",
        ]
    )
    assert result.ret == EXIT_OK

    # Mock assert the correct variables are set



from types import GeneratorType

import pytest

from conftest import ALL_HOSTS

from pytest_ansible.results import ModuleResult


invalid_hosts = ["none", "all", "*", "local*"]


@pytest.fixture()
def adhoc_result(hosts):
    return hosts.all.ping()


def test_len(adhoc_result):
    assert len(adhoc_result) == len(ALL_HOSTS)


def test_keys(adhoc_result):
    assert set(adhoc_result) == set(ALL_HOSTS)


def test_items(adhoc_result):
    items = adhoc_result.items()
    assert isinstance(items, GeneratorType)
    count = 0
    for count, item in enumerate(items, 1):
        assert isinstance(item, tuple)
        assert isinstance(item[0], str)
        assert isinstance(item[1], ModuleResult)
    assert count == len(ALL_HOSTS)


def test_values(adhoc_result):
    values = adhoc_result.values()
    assert isinstance(values, list)
    # assure that it is a copy
    assert values is not adhoc_result.contacted.values()
    count = 0
    for count, val in enumerate(values, 1):
        assert isinstance(val, ModuleResult)
    assert count == len(ALL_HOSTS)


@pytest.mark.parametrize("host", ALL_HOSTS)
def test_contains(adhoc_result, host):
    assert host in adhoc_result


@pytest.mark.parametrize("host", invalid_hosts)
def test_not_contains(adhoc_result, host):
    assert host not in adhoc_result


@pytest.mark.parametrize("host_pattern", ALL_HOSTS)
def test_getitem(adhoc_result, host_pattern):
    assert adhoc_result[host_pattern]
    assert isinstance(adhoc_result[host_pattern], ModuleResult)


@pytest.mark.parametrize("host_pattern", invalid_hosts)
def test_not_getitem(adhoc_result, host_pattern):
    with pytest.raises(KeyError):
        assert adhoc_result[host_pattern]


@pytest.mark.parametrize("host_pattern", ALL_HOSTS)
def test_getattr(adhoc_result, host_pattern):
    assert hasattr(adhoc_result, host_pattern)
    assert isinstance(adhoc_result[host_pattern], ModuleResult)


@pytest.mark.parametrize("host_pattern", invalid_hosts)
def test_not_getattr(adhoc_result, host_pattern):
    assert not hasattr(adhoc_result, host_pattern)
    with pytest.raises(AttributeError):
        getattr(adhoc_result, host_pattern)


@pytest.mark.requires_ansible_v1()
def test_connection_failure_v1():
    from pytest_ansible.errors import AnsibleConnectionFailure
    from pytest_ansible.host_manager import get_host_manager

    hosts = get_host_manager(inventory="unknown.example.com,", connection="smart")
    with pytest.raises(AnsibleConnectionFailure) as exc_info:
        hosts.all.ping()
    # Assert message
    assert exc_info.value.message == "Host unreachable"
    # Assert contacted
    assert exc_info.value.contacted == {}
    # Assert dark
    assert "unknown.example.com" in exc_info.value.dark
    # Assert unreachable
    assert "failed" in exc_info.value.dark["unknown.example.com"]
    assert exc_info.value.dark["unknown.example.com"]["failed"]
    # Assert msg
    assert "msg" in exc_info.value.dark["unknown.example.com"]
    assert exc_info.value.dark["unknown.example.com"]["msg"].startswith(
        "SSH Error: ssh: Could not resolve hostname" + " unknown.example.com:"
    )


@pytest.mark.requires_ansible_v2()
def test_connection_failure_v2():
    from pytest_ansible.errors import AnsibleConnectionFailure
    from pytest_ansible.host_manager import get_host_manager

    hosts = get_host_manager(inventory="unknown.example.com,", connection="smart")
    with pytest.raises(AnsibleConnectionFailure) as exc_info:
        hosts.all.ping()
    # Assert message
    assert exc_info.value.message == "Host unreachable in the inventory"
    # Assert contacted
    assert exc_info.value.contacted == {}
    # Assert dark
    assert "unknown.example.com" in exc_info.value.dark
    # Assert unreachable
    assert (
        "unreachable" in exc_info.value.dark["unknown.example.com"]
    ), exc_info.value.dark.keys()
    assert exc_info.value.dark["unknown.example.com"]["unreachable"]
    # Assert msg
    assert "msg" in exc_info.value.dark["unknown.example.com"]
    assert (
        "Failed to connect to the host via ssh"
        in exc_info.value.dark["unknown.example.com"]["msg"]
    )


@pytest.mark.requires_ansible_v2()
def test_connection_failure_extra_inventory_v2():
    from pytest_ansible.errors import AnsibleConnectionFailure
    from pytest_ansible.host_manager import get_host_manager

    hosts = get_host_manager(
        inventory="localhost", extra_inventory="unknown.example.extra.com,"
    )
    with pytest.raises(AnsibleConnectionFailure) as exc_info:
        hosts.all.ping()
    # Assert message
    assert exc_info.value.message == "Host unreachable in the extra inventory"
    # Assert contacted
    assert exc_info.value.contacted == {}
    # Assert dark
    assert "unknown.example.extra.com" in exc_info.value.dark
    # Assert unreachable
    assert (
        "unreachable" in exc_info.value.dark["unknown.example.extra.com"]
    ), exc_info.value.dark.keys()
    assert exc_info.value.dark["unknown.example.extra.com"]["unreachable"]
    # Assert msg
    assert "msg" in exc_info.value.dark["unknown.example.extra.com"]
    assert (
        "Failed to connect to the host via ssh"
        in exc_info.value.dark["unknown.example.extra.com"]["msg"]
    )
